/*********************************************************************
 Copyright (C) 2003 Imperial College London

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

/*
 * hive.cc 
 * Video processing software / shot boundary detection
 * Marcus Pickering / Stefan Rueger
 * mjp3@doc.ic.ac.uk
 * June 2003
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <getopt.h>

#include "decoder.h"
#include "mpeg_decoder.h"
#include "sbd_vidprocess.h"
#include "encoder.h"
#include "sbdxml_encoder.h"
#include "ansesvidseg_encoder.h"

const char* program_name;
VidProcess *vprocess;
Decoder *decdr;
Encoder *ecdr;

void print_usage(FILE* stream, int exit_code)
{
  fprintf(stream, "Usage: %s options input_file\n", program_name);
  fprintf(stream,
	  "  -h --help               Display this usage information.\n"
	  "  -o --output-file  file  XML output file.\n"
	  "  -k --keyframe-dir dir   Output directory for keyframes.\n"
	  "  -d --dn-logfile file    Optional file for logging dn values.\n"
	  "  -e --encoder-mode mode  Output format. Mode is\n"
	  "                           1 - XML output\n"
	  "                           2 - ANSES vidseg output\n"
	  "  -m --main-thresh value  Main threshold setting.\n"
	  "  -l --low-thresh value   Gradual threshold setting.\n");
  exit(exit_code);
}

void finishup(int enc_mode)
{
  vprocess->EndOfStream();
  while (vprocess->IsFrameReady())
    {
      frame* gotframe = vprocess->GetFrame();
      switch(enc_mode)
	{
	case 1:
	  //not sure why this can't be handled by ecdr
	  if (gotframe->transType == CUT)
	    printf("<trans type=\"CUT\" preFNum=\"%i\" postFNum=\"%i\"/>\n", gotframe->transStartFr, gotframe->m_iFrameNo);
	  else if (gotframe->transType == GRADUAL)
	    printf("<trans type=\"GRAD\" preFNum=\"%i\" postFNum=\"%i\"/>\n", gotframe->transStartFr, gotframe->m_iFrameNo);
	  break;
	case 2:
	  ecdr->OutputFrame(gotframe);
	  break;
	}
      vprocess->RemoveFrame();
    }
  delete decdr;
  delete vprocess;
  delete ecdr;
  exit(0);
}

void Fatal(const char *fmt,...)
{
   va_list ap; /*will point to each unnamed argument in turn*/
   const char *p=fmt;
   va_start(ap,fmt); /* point to first element after fmt*/

   fprintf(stderr, "FATAL ERROR: ");
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n");

   va_end(ap);
   exit(1);
}

int hive_main(int argc, char* argv[])
{
  program_name = argv[0];

  if (argc == 1)
    print_usage(stdout, 1);

  const char* const short_options = "hk:o:d:e:m:l:";
  const struct option long_options[] = 
  { 
    { "help", 0, NULL, 'h' }, 
    { "keyframe-dir", 1, NULL, 'k' }, 
    { "output-file", 1, NULL, 'o' }, 
    { "dn-logfile", 1, NULL, 'd' }, 
    { "encoder-mode", 1, NULL, 'e' }, 
    { "main-thresh", 1, NULL, 'm' },
    { "low-thresh", 1, NULL, 'l' },  
    { NULL, 0, NULL, 0 }
  };
  
  char* output_filename = NULL;
  char* vid_filename = NULL;
  char* keyframe_dir = NULL;
  char* dnlog_filename = NULL;
  int encoder_mode = 1;
  double main_thresh=0.5, low_thresh=0.05;
  int next_option;
#ifdef _DEBUG
  printf("Processing options\n");
#endif

  do
    {
      next_option = getopt_long(argc, argv, short_options, long_options, NULL);
      
      switch(next_option)
	{
	case 'h' :
	  print_usage(stdout, 0);
	case 'o' :
	  output_filename = optarg;
	  break;
	case 'k' :
	  keyframe_dir = optarg;
	  break;
	case 'd' :
	  dnlog_filename = optarg;
	  break;
	case 'e' :
	  encoder_mode = atoi(optarg);
	  break;
	case 'm' :
	  main_thresh = atof(optarg);
	  break;
	case 'l' :
	  low_thresh = atof(optarg);
	  break;
	case '?' : // Invalid option specified
	  print_usage(stderr, 1);
	case -1 :
	  break;
	default : // How did we get here?!
	  Fatal("Unable to process command line options");
	}
    }while(next_option != -1);
#ifdef _DEBUG
  printf("Done options\n");
#endif

  if (optind >= argc)
    Fatal("Bad command line - maybe no input video specified?");
  
  vid_filename = argv[optind];
#ifdef _DEBUG
  printf("Inititalsising decoder\n");
#endif
  decdr = new FFMPEG_Decoder(vid_filename);
#ifdef _DEBUG
  printf("Done decoder\n");
#endif

  /*
  VideoProperties *vprop = decdr->GetVideoProperties();
#ifdef _DEBUG
  printf("Done properties\n");
#endif
  */

  vprocess = new SBD_VidProcess(main_thresh, low_thresh, dnlog_filename);
#ifdef _DEBUG
  printf("Done new vidprocess\n");
#endif

  if (!output_filename) {
    Fatal("XML output filename must be specified with option -o");
  }

  switch(encoder_mode)
    {
    case 1 :
      ecdr = new SBDXML_Encoder(output_filename, vid_filename, keyframe_dir);
      break;
    case 2 :
      ecdr = new AnsesVidseg_Encoder(output_filename, vid_filename, keyframe_dir);
      break;
    default :
      Fatal("Encoder mode error");
      break;
    }

  frame* vidframe;
  while (!decdr->EndOfVideo())
    {

      if (decdr->GetFrame(vidframe))
	break;
#ifdef _DEBUG
      printf("  Process frame\n");
#endif
      vprocess->ProcessFrame(vidframe);

#ifdef _DEBUG
      printf("while loop\n");
#endif

      //We don't care about audio when we're just outputting frame nos.
      while (vprocess->IsFrameReady())
	{

#ifdef _DEBUG
	  printf("  output frame\n");
#endif
	  ecdr->OutputFrame(vprocess->GetFrame());
#ifdef _DEBUG
	  printf("  remove frame\n");
#endif
	  vprocess->RemoveFrame();

#ifdef _DEBUG
	  printf("  done\n");
#endif

	}

      long frame_num = decdr->FrameNumber();
      fprintf(stderr, "Frame: %d\r", frame_num);

#ifdef _DEBUG
      printf("loop end\n");
#endif

    }

  finishup(encoder_mode);

  return 0;
}











