#include <math.h>
#include <stdio.h>
#include "sbd_vidprocess.h"

extern void Fatal(const char *fmt,...);

SBD_VidProcess::SBD_VidProcess(double mthresh, double lthresh, 
			       char *dnlogfilename)
{
  mainthresh = mthresh;
  gradthresh = lthresh;
  D8thresh = 2.0*mainthresh / 5.0;
  transPending = 0;
  transStart = 0;
  transSavedStart = 0;
  prevEnd = 0; 
  isEndOfStream = 0;
  shotNo = 0;
  fdnLog = NULL;
  //set up dn log file if requested
  if (dnlogfilename != NULL)
    {
      if ((fdnLog = fopen(dnlogfilename, "w")) == NULL)
	Fatal("Failed to open dn log file \"%s\" for writing", dnlogfilename);
    }
}

SBD_VidProcess::~SBD_VidProcess()
{
  if (fdnLog)
    fclose(fdnLog);
}

void SBD_VidProcess::ProcessFrame(frame* vidframe)
{
  //printf("PROCESS: start\n");
  int frameno = vidframe->m_iFrameNo;
  
  //Flag the first key frame
  if (frameno == 10)
    vidframe->isKeyFrame = 1;

  frameQ.push_back(*vidframe);
  CreateHistogram(vidframe);

  //printf("PROCESS: histogram done\n");
  if (frameno >= 16) {
    CalculateAllDistances((int)frameQ.size()-1);    
  }

  //printf("PROCESS: distances done\n");
  if (frameno >= 31)
    {
      int ref = frameQ.size()-16; //15 frames back from last frame
      //printf("PROCESS: calculating dn\n");
      CalculateDn(ref); 
      //printf("PROCESS: dn done\n");
      //printf("no: %i, d16: %f, d8: %f, d4: %f, d2: %f\n", frameQ[ref].m_iFrameNo, histQ[ref].dn[3], histQ[ref].dn[2],histQ[ref].dn[1], histQ[ref].dn[0]);
      double currD4 = histQ[ref].dn[1];
      double prevD4 = (histQ.size() == 32) ? 0 : (histQ[ref-1].dn[1]);

      //Flag potential start
      if ((currD4 >= gradthresh) && (prevD4 < gradthresh))
	frameQ[ref-1].isTransStart = 1;
    
      if ((currD4 < gradthresh) && (prevD4 >= gradthresh)) 
	{
	  //A previously declared gradual transition has come to an end.
	  //The following condition ensures that two peaks
	  //  within a transition don't cause it to be declared twice
	  if ((transStart >= prevEnd) && transPending)
	    {
	      frameQ[ref].transType = GRADUAL;
	      prevEnd = frameQ[ref].m_iFrameNo;
	      transPending = 0;
	    }
	  frameQ[ref].isTransEnd = 1;
	  frameQ[ref].transStartFr = transStart;
	}
    }
  if ((frameno >= 63) && !transPending)
    {
      int ref = frameQ.size()-32;

      if (frameQ[ref].m_iFrameNo == (prevEnd + 10))
	frameQ[ref].isKeyFrame = 1;
      //printf("PROCESS: check for shot change\n");
      transition trans = CheckForShotChange(ref);
      if ((trans == CUT) && (frameQ[ref].m_iFrameNo > prevEnd))
	{
	  frameQ[ref].isTransEnd = 1;
	  frameQ[ref].transStartFr = frameQ[ref].m_iFrameNo - 1;
	  frameQ[ref].transType = trans;
	  prevEnd = frameQ[ref].m_iFrameNo;
	  shotNo++;
	}
      else if ((trans == GRADUAL) && (frameQ[ref].m_iFrameNo > prevEnd))
	{
	  //printf("Gradual noticed at %i\n", frameQ[ref].m_iFrameNo);
	  //Look for its start
	  //If the start isn't found in the buffer, then it's the saved one 
	  //  from ages ago.
	  transStart = transSavedStart;
	  for (int i=ref-1; i>=0; i--)
	    if (frameQ[i].isTransStart)
	      {
		transStart = frameQ[i].m_iFrameNo;
		break;
	      }
	  //Look for its end
	  //If the end isn't in the buffer, set the pending flag and finish
	  //  it later.
	  if (transStart > prevEnd)
	    {
	      transPending = 1;
	      for (int i=ref; i<frameQ.size(); i++)
		if (frameQ[i].isTransEnd)
		  {
		    transPending=0;
		    frameQ[i].transType = GRADUAL;
		    frameQ[i].transStartFr = transStart;
		    prevEnd = frameQ[i].m_iFrameNo;
		    break;
		  }
	      shotNo++;
	    }
	}
    }
  //printf("PROCESS: done\n");
}

void SBD_VidProcess::EndOfStream()
{
  isEndOfStream = 1; //redundant?
}

int SBD_VidProcess::IsFrameReady()
{
  return ((frameQ.size() >= 48) || (isEndOfStream && frameQ.size()));
}

frame* SBD_VidProcess::GetFrame()
{
  return &frameQ.front();
}

void SBD_VidProcess::RemoveFrame()
{
  if (frameQ.front().isTransStart)
    transSavedStart = frameQ.front().m_iFrameNo;

  frameQ.pop_front();
  histQ.pop_front();
}

void SBD_VidProcess::CreateHistogram(frame *vidframe)
{
  Histogram theHistogram;
  memset(&theHistogram.hist, 0, (3*sizeof(int)*HIST_BLOCKS*HIST_BINS));
  memset(&theHistogram.blkc, 0, HIST_BLOCKS*3*sizeof(int));

  int iSize = vidframe->m_iHeight * vidframe->m_iWidth;

  unsigned char *imgPtr = vidframe->m_pcImage;
  
  for (int com=0; com<3; com++)
    {
      int iBlockWidth =(int)( vidframe->m_iWidth / (sqrt(HIST_BLOCKS)));
      int iBlockHeight = (int)(vidframe->m_iHeight / (sqrt(HIST_BLOCKS)));
      int iImgSize = iSize;
      int iImgWidth = vidframe->m_iWidth;
      if (com)
	{
	  //U and V components are subsampled by 2
	  iBlockWidth /= 2;
	  iBlockHeight /= 2;
	  iImgWidth /= 2;
	  iImgSize /= 4;
	}
      for (int i=0; i<iImgSize; i++)
	{
	  int x = i%iImgWidth;
	  int y = (i-x) / iImgWidth;

	  //Determine which block the pixel falls in
	  //This isn't very general at the moment (specific to 9 blocks)
	  int iBlockNo;
	  if (x >= 2*iBlockWidth)
	    {
	      if (y >= 2*iBlockHeight)
		iBlockNo = 8;
	      else if (y >= iBlockHeight)
		iBlockNo = 5;
	      else
		iBlockNo = 2;
	    }
	  else if (x >= iBlockWidth)
	    {
	      if (y >= 2*iBlockHeight)
		iBlockNo = 7;
	      else if (y >= iBlockHeight)
		iBlockNo = 4;
	      else
		iBlockNo = 1;
	    }
	  else
	    {
	      if (y >= 2*iBlockHeight)
		iBlockNo = 6;
	      else if (y >= iBlockHeight)
		iBlockNo = 3;
	      else
		iBlockNo = 0;
	    }
	  int iBin = (int)(HIST_BINS * (double)(*(imgPtr++)) / 256.0);

	  if (iBin < 0)
	    cerr << "iBin " << iBin;
	  (theHistogram.hist[com][iBlockNo][iBin])++;
	  theHistogram.blkc[com][iBlockNo]++;
	}
    }
  histQ.push_back(theHistogram);

}

void SBD_VidProcess::CalculateAllDistances(int ref)
{
  CalculateDistance(ref, 2);
  CalculateDistance(ref, 4);
  CalculateDistance(ref, 8);
  CalculateDistance(ref, 16);
}

void SBD_VidProcess::CalculateDistance(int ref, int offset)
{
  double pdDistances[HIST_BLOCKS];

  for (int i=0; i<HIST_BLOCKS; i++)
    {
      int iYsum=0, iUsum=0, iVsum=0;
      for (int j=0; j<HIST_BINS; j++)
	{
	   iYsum += abs((int)((histQ[ref].hist)[0][i][j]
				- (histQ[ref-offset].hist)[0][i][j]));
	   iUsum += abs((int)((histQ[ref].hist)[1][i][j]
			      - (histQ[ref-offset].hist)[1][i][j]));
	   iVsum += abs((int)((histQ[ref].hist)[2][i][j]
			      - (histQ[ref-offset].hist)[2][i][j]));
	}
      double dYnorm = iYsum / (double)(histQ[ref].blkc)[0][i];
      double dUnorm = iUsum / (double)(histQ[ref].blkc)[1][i];
      double dVnorm = iVsum / (double)(histQ[ref].blkc)[2][i];

      pdDistances[i] = (double)MAX(dYnorm, MAX(dUnorm, dVnorm)); 
    }
  //Find median of block values
  int iMedianLim = (int)ceil((double)HIST_BLOCKS / 2.0);
  for (int d=0; d<iMedianLim; d++)
    for(int blk=d+1; blk<HIST_BLOCKS; blk++)
      if (pdDistances[blk] < pdDistances[d])
	{
	  //swap them
	  double temp = pdDistances[blk];
	  pdDistances[blk] = pdDistances[d];
	  pdDistances[d] = temp;
	}

  int iDistIndex;
  switch(offset)
    {
    case 2:
      iDistIndex = 0;
      break;
    case 4:
      iDistIndex = 1;
      break;
    case 8:
      iDistIndex = 2;
      break;
    case 16:
      iDistIndex = 3;
      break;
    default:
      cerr << "* * * Error in diff referencing * * *" << endl;
    }

  histQ[ref].distances[iDistIndex] = pdDistances[iMedianLim-1];
}

void SBD_VidProcess::CalculateDn(int ref)
{
  //calculate d2,d4,d8,d16
  //printf("DN: start\n");
  for (int i=0; i<4; i++)
    {
      //printf("DN: %i\n", i);
      int n = (int)pow(2, i+1);

      double dDn = 0;
      //printf("DN: distances\n");
      for (int j=0; j<n; j++)
	dDn += histQ[ref+j].distances[i];

      dDn /= (double)n;
      //printf("DN: hist\n");
      histQ[ref].dn[i] = dDn;
      //printf("DN: log\n");
      //printf("D%i is %f\n", n, dDn);
      if (fdnLog)
	fprintf(fdnLog, "%f ", dDn);
      //printf("DN: log done\n");
    }
  if (fdnLog)
    fprintf(fdnLog, "\n");
  //printf("DN: done\n");
}

transition SBD_VidProcess::CheckForShotChange(int ref)
{
  transition res = CheckForCut(ref);
  if (res == NONE)
    res = CheckForGrad(ref);

  return res;
}

transition SBD_VidProcess::CheckForCut(int ref)
{
  /*
   * Current criteria for a cut:
   *   - dD16 gt threshold (NB d16 not involved in any other way)
   *   - dD8 gt threshold 
   *   - dD8, dD4, dD2 gt 2 neighbours either side
   *   - all dDn gt (1.3 * dD(n/2)) 
   */
  double dD16 = histQ[ref].dn[3];
  double dD8 = histQ[ref].dn[2];
  double dD4 = histQ[ref].dn[1];
  double dD2 = histQ[ref].dn[0];
  
  if (dD8 < D8thresh)
    return NONE;

  //Glitch detection
  if (dD8 > (1.414*dD16))
    return NONE;

  //coincident peaks?
  for (int i=-2; i<=2; i++)
    {
      if (histQ[ref+i].dn[2] > dD8)
	return NONE;
      if (histQ[ref+i].dn[1] > dD4)
	return NONE;
      if (histQ[ref+i].dn[0] > dD2)
	return NONE;
    }
  
  //peaks close enough in magnitude?
  /*  if ((dD8 * 1.3) < dD16)
      return NONE;
  */

  //Peaks must be close to the d_2n peak
  //sqrt(2) ~= 1.414
  if (((dD4 * 1.414) < dD8) || ((dD2 * 1.414) < dD4))
    return NONE;


  return CUT;
}
transition SBD_VidProcess::CheckForGrad(int ref)
{
  //check for peak in D16
  //and D8 peak within +/- 4 of D16 peak

  double dD16 = histQ[ref].dn[3];
  double dD8 = histQ[ref].dn[2];
  double dD4 = histQ[ref].dn[1];
  double dD2 = histQ[ref].dn[0];

  if (dD16 <= mainthresh)
    {
      return NONE;
    }

  int iD16peakIndex=0, iD8peakIndex=0, iD4peakIndex=0, iD2peakIndex=0;
  //19/06/02 - this doesn't look to be a very efficient way of doing things
  for (int i=-16; i<=16; i++)
    {
      if (histQ[ref+i].dn[3] >dD16)
	{
	  return NONE;
	}
      if (histQ[ref+i].dn[2] >dD8)
	{ 
	  dD8 = histQ[ref+i].dn[2];
	  iD8peakIndex = i;
	}
      //See where the D4 peak is
      if (histQ[ref+i].dn[1] >dD4)
	{
	  dD4 = histQ[ref+i].dn[1];
	  iD4peakIndex = i;
	}
      //See where the D2 peak is
      if (histQ[ref+i].dn[0] >dD2)
	{
	  dD2 = histQ[ref+i].dn[0];
	  iD2peakIndex = i;
	}
    }
  //d8 peak must be within factor of sqrt(2) d16 to be considered coincident
  //Pye et al specify d8 peak within 4, but assume that's 25fps? 
  // I'm going for 5....
  if ((dD8 <= D8thresh) //|| ((dD8 * sqrt(2)) < dD16) 
      || (abs(iD8peakIndex) > 5))
    {
      return NONE;
    }

  return GRADUAL;
}


